/*
 * Copyright (c) 2000, 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Clack.
 * 
 * Clack is free software, also known as "open source."  You can redistribute
 * it and/or modify it under the terms of the GNU General Public License
 * (GPL), version 2, as published by the Free Software Foundation (FSF).
 * 
 * Clack is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with Clack; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 * 
 * Clack is modelled on and includes some code fragments from MIT's Click
 * system (http://www.pdos.lcs.mit.edu/click/) which is copyright (c)
 * 1999-2000 Massachusetts Institute of Technology and also GPL'ed.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <errno.h>
#include <clack/c/packet.h>
#include <clack/c/packet_eth.h>
#include <clack/c/print_packet.h>
#include <clack/c/if_arp.h>
#include <arpa/inet.h>
#include <clack/c/arp_responder.h>

/*
 * Imported:  push_in1 - Output link for Ether/IP and Ether/Arp packets.
 *            push_in2 - Output link for failed arp lookups.
 *            ethaddr  - Our ethernet address.
 *	      ipaddr   - Our parsed IP address.
 */
extern clack_error_t	push_in1(clack_packet_t *p);
extern clack_error_t	push_in2(clack_packet_t *p);
extern unsigned char	my_ethaddr[ETHER_ADDR_SIZE];
extern struct in_addr   my_ipaddr;

/*
 * Exported: push_out1 - Input link for IP packets to be sent out.
 *           push_out2 - Input link for Arp responses we sent out.
 */
clack_error_t		push_out1(clack_packet_t *p);
clack_error_t		push_out2(clack_packet_t *p);

/*
 * Is this correct?
 */
static unsigned char	bcastaddr[] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

/*
 * Locally allocated table. 
 */
struct arp_mapping_entry {
	struct in_addr	am_ip;
	unsigned char	am_eth[ETHER_ADDR_SIZE];
	char		valid;
	char		inuse;
	clack_packet_t  *packet;
};

/*
 * Statically sized, linear search. XXX IMPROVE LATER!
 */
#define MYSIZE	100
static struct arp_mapping_entry		arp_mapping_table[MYSIZE];

/*
 * Initialize
 */
clack_error_t
init(void) 
{
	printf("arp_querier(init):\n");

	return 0;
}

/*
 * We get outgoing IP packets on this input link.
 */
clack_error_t
push_out1(clack_packet_t *p)
{
	int			   i;
	struct arp_mapping_entry  *mapping;
	struct ether_arp	  *r;
	clack_packet_t		  *request;
	
	/*
	 * Search the table for a valid entry with a matching IP address.
	 */
	for (i = 0; i < MYSIZE; i++) {
		mapping = &arp_mapping_table[i];

		if (mapping->inuse &&
		    mapping->am_ip.s_addr == p->anno.next_hop.s_addr) {
			/*
			 * If the mapping is not valid yet, must break out.
			 */
			if (!mapping->valid)
				break;
			
			add_eth_hdr(p, mapping->am_eth,
				    my_ethaddr, ETHERTYPE_IP);
			push_in1(p);
			return 0;
		}
	}

	/*
	 * Is there a matching, but not yet valid entry?
	 */
	if (i < MYSIZE && !mapping->valid) {
		/*
		 * For now, just drop the packet. We only queue a single
		 * packet.
		 */
		assert(mapping->packet);
		printf("arp_querier(push_out1): Dropping packet\n");
		clack_packet_release(p);          
		return 0;
	}

	/*
	 * No matching entry. Allocate one.
	 */
	for (i = 0; i < MYSIZE; i++) {
		mapping = &arp_mapping_table[i];

		if (! mapping->inuse)
			break;
	}

	/*
	 * No more entries. Just fold for now.
	 */
	if (i == MYSIZE) {
		panic("arp_querier(push_out1): No more table entries!\n");
	}

	/*
	 * Initialize the entry. Note that we save up the original packet.
	 * TODO: The saved packet needs to be a queue of saved packets.
	 */
	mapping->inuse  = 1;
	mapping->valid  = 0;
	mapping->packet = p;
	mapping->am_ip.s_addr = p->anno.next_hop.s_addr;
	
	/*
	 * Form up an arp request packet and send it.
	 */
	request = clack_packet_alloc_padded(18+sizeof (struct ether_arp),
					    sizeof (struct ether_header));

	r = (struct ether_arp *) clack_packet_data(request);
	r->ea_hdr.ar_hrd = htons(ARPHRD_ETHER);
	r->ea_hdr.ar_pro = htons(ETHERTYPE_IP);
	r->ea_hdr.ar_hln = ETHER_ADDR_SIZE;
	r->ea_hdr.ar_pln = sizeof (struct in_addr);
	r->ea_hdr.ar_op  = htons(ARPOP_REQUEST);

	/* Him */
	memset(r->arp_tha, 0, ETHER_ADDR_SIZE);
	memcpy(r->arp_tpa, &p->anno.next_hop, sizeof (struct in_addr));
	/* Me */
	memcpy(r->arp_sha, my_ethaddr, ETHER_ADDR_SIZE);
	memcpy(r->arp_spa, &my_ipaddr, sizeof (struct in_addr));
	/* Goes to the broadcast address on the outgoing interface. */
	add_eth_hdr(request, bcastaddr, my_ethaddr, ETHERTYPE_ARP);
	push_in1(request);
	return 0;
}

/*
 * We get ARP repsonses on this link.
 */
clack_error_t
push_out2(clack_packet_t *p)
{
	struct ether_header *e  = (struct ether_header *) clack_packet_data(p);
	struct ether_arp    *ea = (struct ether_arp *) (e + 1);
	struct arp_mapping_entry  *mapping;
	int			  i;

#if 1
        print_ether_header(e);
        printf(" : ");
        print_ether_ARP(ea);
        printf("\n");
#endif
	
	/*
	 * Search the table for a valid entry with a matching IP address.
	 */
	for (i = 0; i < MYSIZE; i++) {
		mapping = &arp_mapping_table[i];

		if (mapping->inuse && !mapping->valid &&
		    mapping->am_ip.s_addr ==
		    ((struct in_addr *) ea->arp_spa)->s_addr)
			break;
	}
	if (i == MYSIZE) {
		printf("arp_querier(push_out2): Duplicate ARP info!\n");
		return 0;
	}

	/*
	 * Validate the entry
	 */
	memcpy(mapping->am_eth, ea->arp_sha, ETHER_ADDR_SIZE);
	mapping->valid = 1;

	/*
	 * If a packet is queued, push it out.
	 */
	if ((p = mapping->packet)) {
		mapping->packet = 0;
		add_eth_hdr(p, mapping->am_eth, my_ethaddr, ETHERTYPE_IP);
		push_in1(p);
	}
	
	return 0;
}

